import logging
import shutil
from pathlib import Path
from typing import Dict, Optional, Union

import pkg_resources
from jinja2 import Environment, PackageLoader, select_autoescape
from slugify import slugify

import nprompter
from nprompter.api.notion_client import NotionClient


class HtmlNotionProcessor:
    def __init__(
        self, notion_client: NotionClient, output_folder: Union[str, Path], configuration: Optional[Dict] = None
    ):
        self.notion_client = notion_client
        self.output_folder = Path(output_folder)
        self.env = Environment(
            loader=PackageLoader("nprompter", package_path="web/templates"), autoescape=select_autoescape()
        )
        self.assets_folder = Path(pkg_resources.resource_filename("nprompter", "web/assets/"))
        self.script_template = self.env.get_template("script.html")
        self.index_template = self.env.get_template("index.html")
        self.logger = logging.getLogger("NotionProcessor")
        self.configuration = configuration or {}
        self.custom_css = []

    def prepare_folder(self, configuration: Dict):
        if not self.output_folder.exists():
            self.output_folder.mkdir(parents=True)

        js_template = self.env.get_template("script.js")
        css_template = self.env.get_template("nprompter.css")

        with open(self.output_folder / "script.js", "w", encoding="utf8") as writeable:
            writeable.write(js_template.render(**configuration))

        with open(self.output_folder / "nprompter.css", "w", encoding="utf8") as writeable:
            writeable.write(css_template.render(**configuration))

        shutil.copytree(self.assets_folder, self.output_folder, dirs_exist_ok=True)

    def add_extra_style(self, path: Path):
        self.custom_css.append(path.name)
        shutil.copy(path, self.output_folder)

    def process_databases(self, database_id: str, property_filter: str, property_value: str):
        db = self._process_single_database(database_id, property_filter, property_value)

        content = self.index_template.render(databases=[db], version=nprompter.__version__, custom_css=self.custom_css)
        with open(self.output_folder / "index.html", "w", encoding="utf8") as writeable:
            writeable.write(content)

    def _process_single_database(self, database_id: str, property_filter: str, property_value: str):
        database = self.notion_client.get_database(database_id)
        pages = self.notion_client.get_pages(database_id, property_filter, property_value)
        # Create database folder
        (self.output_folder / database_id).mkdir(exist_ok=True)
        database_dict = {"title": database["title"][0]["plain_text"], "scripts": []}
        for page in pages:
            database_dict["scripts"].append(self.process_page(database_id, page))
        return database_dict

    def process_page(self, database_id: str, page: dict):
        title = page["properties"]["Name"]["title"][0]["text"]["content"]
        title_slug = slugify(title)
        blocks = self.notion_client.get_blocks(page["id"])
        block_contents = self.process_blocks(blocks)
        content = self.script_template.render(
            elements=block_contents, title=title, version=nprompter.__version__, custom_css=self.custom_css
        )

        file_name = Path(self.output_folder, database_id, f"{title_slug}.html")
        with open(file_name, "w", encoding="utf8") as writeable:
            writeable.write(content)

        from_root_path = f"{database_id}/{title_slug}.html"
        return {"title": title, "path": from_root_path}

    def process_blocks(self, blocks):
        block_contents = []

        bulleted_list_item = False
        numbered_list_item = False
        for block in blocks:
            block_type = block["type"]

            # Control for lists
            if block_type == "bulleted_list_item" and not bulleted_list_item:
                block_contents.append("<ul>")
                bulleted_list_item = True
            elif block_type != "bulleted_list_item" and bulleted_list_item:
                block_contents.append("</ul>")

            if block_type == "numbered_list_item" and not numbered_list_item:
                block_contents.append("<ol>")
                numbered_list_item = True
            elif block_type != "numbered_list_item" and numbered_list_item:
                block_contents.append("</ol>")

            if block_type == "paragraph":
                if data := self.process_paragraph(block, "paragraph", "p"):
                    block_contents.append(data)
            elif block_type == "quote" or block_type == "callout":
                if data := self.process_paragraph(block, block_type, "blockquote"):
                    block_contents.append(data)
            elif block_type == "bulleted_list_item" or block_type == "numbered_list_item":
                if data := self.process_paragraph(block, block_type, "li"):
                    block_contents.append(data)
            elif block_type.startswith("heading_"):
                size = block_type[-1]
                if data := self.process_paragraph(block, block_type, f"h{size}"):
                    block_contents.append(data)
            elif block_type == "equation":
                expression = block["equation"]["expression"]
                block_contents.append(f"<p>${expression}$</p>")
            elif block_type == "divider":
                if self.configuration["processor"]["skip_on_break"]:
                    self.logger.info("Found divider, stopping block processing")
                    break
                else:
                    block_contents.append("<hr />")
            else:
                block_contents.append(f"<p>⚠ {block['type']} ⚠</p>")
                block_contents.append(f"<!-- Block of type {block['type']} is not currently supported by Nprompter -->")
                self.logger.warning(f"Block of type {block['type']} is not currently supported by Nprompter")

        return block_contents

    def process_paragraph(self, block, block_type, tag_name):
        notion_color = block[block_type].get("color", "default")
        base_classes = [block_type, f"notion-{notion_color}"]
        contents = block[block_type].get("text", block[block_type].get("rich_text", []))
        paragraph_content_tags = []
        for content in contents:
            if text := content.get("text"):
                text_content = text["content"].replace("\n", "<br />")
                annotations = content["annotations"]
                annotations_tags = ["bold", "italic", "strikethrough", "underline", "code"]
                classes = " ".join(base_classes + [tag for tag in annotations_tags if annotations.get(tag)])
                tag = f'<span class="{classes}">{text_content}</span>'
                paragraph_content_tags.append(tag)
            elif equation := content.get("equation"):
                paragraph_content_tags.append(f'${equation["expression"]}$')
        if paragraph_content_tags:
            paragraph_content = "".join(paragraph_content_tags)
            return f"<{tag_name}>{paragraph_content}</{tag_name}>"
        return None



from typing import Dict, List, Tuple

import requests


def build_filter_query(properties: Tuple[str, str, str]) -> Dict:
    """
    Builds a query to filter for pages with the properties specified as triplets

    :param properties: A list of triplets containing (property, condition, value)
    :return:
    """

    or_conditions = [
        {"property": property_name, "status": {condition: value}} for property_name, condition, value in properties
    ]

    return {"filter": {"or": or_conditions}}


class NotionClient:
    def __init__(self, notion_api_key: str, notion_version: str):
        self.headers = {
            "Authorization": f"Bearer {notion_api_key}",
            "Notion-Version": notion_version,
        }

    def get_database(self, database_id: str) -> Dict:
        database = requests.get(
            f"https://api.notion.com/v1/databases/{database_id}",
            headers=self.headers,
        ).json()
        return database

    def get_pages(self, database_id: str, property_filter: str, property_value: str) -> List[Dict]:
        query = build_filter_query([(property_filter, "equals", property_value)])
        database = requests.post(
            f"https://api.notion.com/v1/databases/{database_id}/query",
            json=query,
            headers=self.headers,
        ).json()

        if database.get("object") == "error":
            raise ValueError(f"{database['code']} - {database['message']}")

        return database["results"]

    def get_blocks(self, page_id: str) -> List[Dict]:
        blocks = []
        page = requests.get(
            f"https://api.notion.com/v1/blocks/{page_id}/children",
            headers=self.headers,
        ).json()
        blocks.extend(page["results"])

        while start_cursor := page.get("next_cursor", None):
            page = requests.get(
                f"https://api.notion.com/v1/blocks/{page_id}/children",
                params={"start_cursor": start_cursor},
                headers=self.headers,
            ).json()
            blocks.extend(page["results"])
        return blocks



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nprompter',
 'nprompter.api',
 'nprompter.cli',
 'nprompter.processing',
 'nprompter.web']

package_data = \
{'': ['*'], 'nprompter.web': ['assets/*', 'templates/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'python-slugify>=6.1.2,<7.0.0',
 'requests>=2.28.0,<3.0.0',
 'tomli>=2.0.1,<3.0.0',
 'typer>=0.4.1,<0.5.0']

entry_points = \
{'console_scripts': ['nprompter = nprompter.__main__:app']}

setup_kwargs = {
    'name': 'nprompter',
    'version': '3.3.0',
    'description': 'A web based teleprompter that uses Notion as a storage backend',
    'long_description': None,
    'author': 'Antonio Feregrino',
    'author_email': 'antonio.feregrino@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import importlib
import os

import tomli

import nprompter
import nprompter.cli.defaults


def get_config(configuration_file):
    config_dict = tomli.loads(importlib.resources.read_text(nprompter.web, "config.toml"))
    if os.path.exists(configuration_file or nprompter.cli.defaults.DEFAULT_APPEARANCE_CONFIG_PATH):
        with open(configuration_file, "rb") as readable:
            user_defined_config = tomli.load(readable)
            config_dict.update(user_defined_config)
    return config_dict



import importlib
import os
import socketserver
import webbrowser
from http.server import SimpleHTTPRequestHandler
from pathlib import Path
from typing import Any, Optional

import typer

import nprompter
import nprompter.web
from nprompter.api.notion_client import NotionClient
from nprompter.cli.defaults import DEFAULT_OUTPUT_PATH, DEFAULT_PORT, DEFAULT_PROPERTY, DEFAULT_VALUE
from nprompter.cli.helpers import get_config
from nprompter.processing.processor import HtmlNotionProcessor

app = typer.Typer(add_completion=False)


@app.command()
def build(
    database_id: str = typer.Argument(..., help="The Notion's database to process"),
    notion_api_key: str = typer.Argument(None, envvar="NOTION_API_KEY", help="The Notion API key to use"),
    output: Optional[Path] = typer.Option(
        DEFAULT_OUTPUT_PATH, "--output", "-o", help="Where the output should be written"
    ),
    property_filter: Optional[str] = typer.Option(
        DEFAULT_PROPERTY, "--filter", "-f", help="The name of the Notion's page property to filter by"
    ),
    property_value: Optional[str] = typer.Option(
        DEFAULT_VALUE, "--value", "-v", help="The value of the Notion's page property to filter by"
    ),
    appearance_file: Optional[Path] = typer.Option(
        None, "--config", "-c", help="A path to an appearance configuration file"
    ),
    custom_css: Optional[Path] = typer.Option(None, "--extra-css", "-s", help="A path to extra css configuration"),
    download_database: bool = typer.Option(
        True, "--download/--no-download", help="Whether to sync with the Notion's database"
    ),
):
    """
    Fetch a Notion database and create the teleprompter's pages from it
    """

    config_dict = get_config(appearance_file)

    notion_version = os.getenv("NOTION_VERSION", nprompter.__notion_version__)

    notion_client = NotionClient(notion_api_key=notion_api_key, notion_version=notion_version)
    processor = HtmlNotionProcessor(notion_client, output_folder=output, configuration=config_dict)

    processor.prepare_folder(config_dict)

    if custom_css:
        processor.add_extra_style(custom_css)

    if download_database:
        processor.process_databases(database_id, property_filter, property_value)


@app.command()
def serve(
    port: int = typer.Argument(DEFAULT_PORT),
    content_directory: Path = typer.Argument(DEFAULT_OUTPUT_PATH),
    browser: bool = typer.Option(True, help="Whether to open a web browser pointing to the server's address"),
):
    """
    Start a basic web server to display the teleprompter's pages.
    This a good option for a small scale deployment.
    """

    class CustomHandler(SimpleHTTPRequestHandler):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, directory=str(content_directory), **kwargs)

        def log_message(self, format: str, *args: Any) -> None:
            pass

    socketserver.TCPServer.allow_reuse_address = True
    with socketserver.TCPServer(("", port), CustomHandler) as httpd:
        location = f"http://localhost:{port}"
        print(f"Serving at {location}")
        if browser:
            webbrowser.open(location)
        httpd.serve_forever()


@app.command()
def create_config(override: bool = False):
    config = importlib.resources.read_text(nprompter.web, "config.toml")

    if not os.path.exists("config.toml") or override:
        with open("config.toml", "w") as writable:
            writable.write(config)
    else:
        print("The file config.toml already exists, call this program with the --override flag to override")
